#include "pwatch_nl.h"

#include <asm/types.h>
#include <sys/socket.h>
#include <linux/genetlink.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>


struct train {
	struct nlmsghdr		n;
	struct genlmsghdr	g;
	char			buf[4096];
};

static int	nl = -1;
static int	pwatch_nl_type = -1;

unsigned int	pwatch_nl_errors = 0;
char		pwatch_nl_strerror[128];

/*
 * Sends a netlink message
 */
static int pwatch_nl_send(__u16 nlmsg_type, __u8 cmd,
	__u16 nla_type,	void *nla_data, int nla_len)
{
	int ret;
	struct train t;
	struct nlattr *attr;
	struct sockaddr_nl sa;
	struct iovec iov;
	struct msghdr msg;
	void *p_nla_data;

	if (nl == -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"nl socket is -1!\n");
		pwatch_nl_errors++;
		return -1;
	}

	memset(&t, 0, sizeof(t));

	attr = (struct nlattr *) ((char *) NLMSG_DATA(&t) + GENL_HDRLEN);
	attr->nla_type = nla_type;
	attr->nla_len = nla_len + NLA_HDRLEN;
	p_nla_data = (void *)((char *)(attr) + NLA_HDRLEN);
	memcpy(p_nla_data, nla_data, nla_len);

	t.n.nlmsg_len = NLMSG_LENGTH(GENL_HDRLEN) + NLMSG_ALIGN(attr->nla_len);
	t.n.nlmsg_type = nlmsg_type;
	t.n.nlmsg_flags = NLM_F_REQUEST;
	t.n.nlmsg_seq = 0;
	t.n.nlmsg_pid = 0;

	t.g.cmd = cmd;
	t.g.version = 1;
	
	iov.iov_base = &t;
	iov.iov_len = t.n.nlmsg_len;

	memset(&sa, 0, sizeof(sa));
	sa.nl_family = AF_NETLINK;

	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (void *) &sa;
	msg.msg_namelen = sizeof(sa);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	ret = sendmsg(nl, &msg, 0);
	if (ret == -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Cannot send msg (%s)", strerror(errno));
		pwatch_nl_errors++;
		return -1;
	}

	return 0;
}

/*
 * Returns type for TASKSTATS.
 */
static int pwatch_nl_get_type(void)
{
	int ret;
	unsigned int len;
	struct train t;
	struct sockaddr_nl sa;
	struct nlattr *attr;
	struct iovec iov;
	struct msghdr msg;

	/* It is cached? */
	if (pwatch_nl_type != -1)
		return pwatch_nl_type;

	if (nl == -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Internal error 1");
		pwatch_nl_errors++;
		return -1;
	}

	if (pwatch_nl_type != -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"internal error2");
		pwatch_nl_errors++;
		return -1;
	}

	/* request */
	ret = pwatch_nl_send(GENL_ID_CTRL, CTRL_CMD_GETFAMILY,
		CTRL_ATTR_FAMILY_NAME, (void *) TASKSTATS_GENL_NAME,
		strlen(TASKSTATS_GENL_NAME) + 1);
	if (ret != 0) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Error sending GETFAMILY command (%s)", 
			strerror(errno));
		pwatch_nl_errors++;
		return -1;
	}

	/* answer */
	memset(&sa, 0, sizeof(sa));
	
	iov.iov_base = (void *) &t;
	iov.iov_len = sizeof(t);
	
	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (void *) &sa;
	msg.msg_namelen = sizeof(sa);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	ret = recvmsg(nl, &msg, 0);
	if (ret == -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Error receiving msg (%s)", strerror(errno));
		pwatch_nl_errors++;
		return -1;
	}

	len = ret;
	if (t.n.nlmsg_type == NLMSG_ERROR) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"NLMSG_ERROR received.");
		pwatch_nl_errors++;
		return -1;
	}

	if (!NLMSG_OK(&t.n, len)) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Corrupt buffer from kernel");
		pwatch_nl_errors++;
		return -1;
	}

	/* Point to first block of data */
	attr = (struct nlattr *) ((char *) NLMSG_DATA(&t) + GENL_HDRLEN);
	/* Skip to second block */
	attr = (struct nlattr *) ((char *) attr + NLA_ALIGN(attr->nla_len));
	if (attr->nla_type != CTRL_ATTR_FAMILY_ID) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"nla_type is wrong (%u). Expected %u.",
			attr->nla_type, CTRL_ATTR_FAMILY_ID);
		pwatch_nl_errors++;
		return -1;
	}

	pwatch_nl_type = *(__u16 *) ((char *) attr + NLA_HDRLEN);

	return pwatch_nl_type;
}

int pwatch_nl_get_taskstats(unsigned int pid, struct taskstats *ts)
{
	int ret, len, type;
	struct train t;
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_nl sa;
	struct nlattr *attr;
	void *p;
	char *block;
	unsigned int aggr_len, aggr_total;
	unsigned int xlen, xtotal;
	unsigned int loaded;

	if (nl == -1) {
		nl = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_GENERIC);
		if (nl == -1) {
			snprintf(pwatch_nl_strerror, 
				sizeof(pwatch_nl_strerror),
				"Cannot create netlink socket (%s)", 
				strerror(errno));
			pwatch_nl_errors++; 
			ret = -1;
			goto out;
		}
	}

	type = pwatch_nl_get_type();
	if (type == -1) {
		ret = -1;
		goto out;
	}

	ret = pwatch_nl_send(type, TASKSTATS_CMD_GET, 
		TASKSTATS_CMD_ATTR_PID, &pid, sizeof(pid));
	if (ret != 0) {
		ret = -2;
		goto out;
	}

	/* receiving */
	memset(&t, 0, sizeof(t));

	iov.iov_base = (void *) &t;
	iov.iov_len = sizeof(t);
	
	memset(&sa, 0, sizeof(sa));
	
	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (void *) &sa;
	msg.msg_namelen = sizeof(sa);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	len = recvmsg(nl, &msg, 0);
	if (len == -1) {
		snprintf(pwatch_nl_strerror, sizeof(pwatch_nl_strerror),
			"Cannot recv msg (%s)", strerror(errno));
		pwatch_nl_errors++;
		ret = -3;
		goto out;
	}

	if (t.n.nlmsg_type == NLMSG_DONE) {
		ret = 0;
		goto out;
	}

	if (t.n.nlmsg_type == NLMSG_ERROR) {
		snprintf(pwatch_nl_strerror, 
			sizeof(pwatch_nl_strerror),
			"Received NLMSG_ERROR from kernel");
		pwatch_nl_errors++;
		ret = -4;
		goto out;
	}

	xtotal = NLMSG_PAYLOAD(&t.n, 0) - GENL_HDRLEN;

	block = (char *) NLMSG_DATA(&t.n) + GENL_HDRLEN;
	attr = (struct nlattr *) block;

	xlen = 0;
	loaded = 0;
	while (xlen < xtotal) {
		xlen += NLA_ALIGN(attr->nla_len);

		if (attr->nla_type == TASKSTATS_TYPE_AGGR_PID) {
			aggr_total = attr->nla_len - NLA_HDRLEN;
			aggr_len = 0;
			attr = (struct nlattr *) ((char *) attr + NLA_HDRLEN);
			while (aggr_len < aggr_total) {
				aggr_len += NLA_ALIGN(attr->nla_len);

				if (attr->nla_type == TASKSTATS_TYPE_STATS) {
					p = (void *) ((char *) attr + NLA_HDRLEN);
					memcpy(ts, p, sizeof(struct taskstats));
					loaded = 1;
					write(sizeof(struct taskstats), p, sizeof(struct taskstats));
				}

				attr = (struct nlattr *) ((char *) attr + aggr_len);
			}
		}

		block += xlen;
		attr = (struct nlattr *) block;
	}

	if (loaded == 0) {
		snprintf(pwatch_nl_strerror, 
			sizeof(pwatch_nl_strerror),
			"No good data from kernel received.");
		pwatch_nl_errors++;
		ret = -5;
		goto out;
	}

	return 0;

	out:
	return ret;
}
